/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/c_gen/c_stub_header_gen.h"

namespace {
#include "idlc/c_gen/c_stub_header_gen_cb.h"
}

namespace tidl {

CStubHeaderGen::CStubHeaderGen(std::shared_ptr<Document> doc)
    : CHeaderGeneratorBase(doc) {}

void CStubHeaderGen::OnInitGen(std::ofstream& stream) {
  GenVersion(stream);
  GenPragmaOnce(stream);
  GenIncludeDefaultHeaders(stream, false);
  GenExplicitLinkageOpen(stream);
  GenStructures(stream);
  GenInterfaces(stream);
}

void CStubHeaderGen::OnFiniGen(std::ofstream& stream) {
  GenExplicitLinkageClose(stream);
}

void CStubHeaderGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;

    const Interface &inf = static_cast<const Interface&>(*i);
    GenInterface(stream, inf);
  }
}

void CStubHeaderGen::GenInterface(std::ofstream& stream,
                                  const Interface& inf) {
  GenInterfaceContext(stream, inf);
  GenInterfaceDelegators(stream, inf);
  GenInterfaceDeclaration(stream, inf);
  GenInterfaceRegister(stream, inf);
  GenInterfaceUnregister(stream, inf);
  GenInterfaceClientNumberGetter(stream, inf);
}

void CStubHeaderGen::GenInterfaceDeclaration(std::ofstream& stream,
                                             const Interface& inf) {
  stream << SmartIndent(GenTemplateString(ReplaceAll(
        CB_INTERFACE_DECL, "##", GetInterfaceIdWithNamespace(inf)),
        [&]()->std::string {
          std::string str;
          for (auto& i : inf.GetDeclarations().GetDecls()) {
            if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
              continue;
            str += NLine(1);
            str += GenTemplateString(CB_INTERFACE_CALLBACK,
                [&]()->std::string {
                  return GetReturnTypeString(i->GetType());
                },
                [&]()->std::string {
                  return i->GetID();
                },
                [&]()->std::string {
                  return GetInterfaceIdWithNamespace(inf);
                },
                [&]()->std::string {
                  std::string s;
                  for (auto& p : i->GetParameters().GetParams()) {
                    if (IsDelegateType(inf, p->GetParameterType()
                        .GetBaseType())) {
                      s += "rpc_port_" + GetInterfaceIdWithNamespace(inf)
                        + "_" + p->GetParameterType().GetBaseType().ToString()
                        + "_h " + p->GetID();
                    } else {
                      s += GetParamTypeString(p->GetParameterType()
                          .GetDirection(), p->GetParameterType().GetBaseType())
                        + p->GetID();
                    }
                    s += ", ";
                  }
                  return s;
                });
          }
          return str;
        }));
}

void CStubHeaderGen::GenInterfaceContext(std::ofstream& stream,
                                         const Interface& inf) {
  GenInterfaceContextDeclaration(stream, inf);
  GenInterfaceContextTagSetter(stream, inf);
  GenInterfaceContextTagGetter(stream, inf);
  GenInterfaceContextSenderGetter(stream, inf);
}

void CStubHeaderGen::GenInterfaceContextDeclaration(
    std::ofstream& stream, const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CONTEXT_DECL, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceContextTagSetter(
    std::ofstream& stream, const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CONTEXT_SET_TAG, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceContextTagGetter(
    std::ofstream& stream, const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CONTEXT_GET_TAG, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceContextSenderGetter(
    std::ofstream& stream, const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CONTEXT_GET_SENDER, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceDelegators(std::ofstream& stream,
                                            const Interface& inf) {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
      continue;
    GenInterfaceDelegator(stream, GetInterfaceIdWithNamespace(inf), *i);
  }
}

void CStubHeaderGen::GenInterfaceDelegator(std::ofstream& stream,
                                           const std::string& id,
                                           const Declaration& decl) {
  GenInterfaceDelegatorDeclaration(stream, id, decl);
  GenInterfaceDelegatorDestructor(stream, id, decl);
  GenInterfaceDelegatorCloner(stream, id, decl);
  GenInterfaceDelegatorInvoker(stream, id, decl);
}

void CStubHeaderGen::GenInterfaceDelegatorDeclaration(
    std::ofstream& stream, const std::string& id, const Declaration& decl) {
  stream << ReplaceAll(
      CB_INTERFACE_DELEGATOR_DECL, "##", id + "_" + decl.GetID());
}

void CStubHeaderGen::GenInterfaceDelegatorDestructor(
    std::ofstream& stream, const std::string& id, const Declaration& decl) {
  stream << ReplaceAll(
      CB_INTERFACE_DELEGATOR_DTOR, "##", id + "_" + decl.GetID());
}

void CStubHeaderGen::GenInterfaceDelegatorCloner(
    std::ofstream& stream, const std::string& id, const Declaration& decl) {
  stream << ReplaceAll(
      CB_INTERFACE_DELEGATOR_CLONER, "##", id + "_" + decl.GetID());
}

void CStubHeaderGen::GenInterfaceDelegatorInvoker(
    std::ofstream& stream, const std::string& id, const Declaration& decl) {
  stream << GenTemplateString(ReplaceAll(
      CB_INTERFACE_DELEGATOR_INVOKER, "##", id + "_" + decl.GetID()),
      [&]()->std::string {
        std::string str;
        for (auto& i : decl.GetParameters().GetParams()) {
          str += ", ";
          str += GetParamTypeString(i->GetParameterType().GetDirection(),
              i->GetParameterType().GetBaseType()) + i->GetID();
        }
        return str;
      });
}

void CStubHeaderGen::GenInterfaceRegister(std::ofstream& stream,
                                          const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_REGISTER, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceUnregister(std::ofstream& stream,
                                            const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_UNREGISTER, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CStubHeaderGen::GenInterfaceClientNumberGetter(std::ofstream& stream,
                                                    const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CLIENT_NUMBER_GETTER, "##",
      GetInterfaceIdWithNamespace(inf));
}

}  // namespace tidl
