/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>

#include <iostream>

#include "idlc/declaration.h"
#include "idlc/type.h"
#include "idlc/parameter.h"

class DeclarationTest : public testing::Test {
 public:
  tidl::Parameters* params;

  virtual void SetUp() {
    params = new tidl::Parameters();
    EXPECT_NE(params, nullptr);
    params->Add(new tidl::Parameter("test1",
          new tidl::ParameterType(new tidl::BaseType("int", "")), __LINE__));
    params->Add(new tidl::Parameter("test2",
          new tidl::ParameterType(new tidl::BaseType("int", "")), __LINE__));
  }
  virtual void TearDown() {}
};

TEST_F(DeclarationTest, Declaration_Constructor) {
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", __LINE__);
  EXPECT_NE(decl, nullptr);
  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetID) {
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", __LINE__);
  EXPECT_NE(decl, nullptr);
  EXPECT_EQ(decl->GetID(), "test");
  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetType) {
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", __LINE__);
  EXPECT_NE(decl, nullptr);
  EXPECT_EQ(decl->GetType().ToString(), "int");
  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetParameters) {
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", __LINE__);
  EXPECT_NE(decl, nullptr);

  int count = 0;
  for (auto& param : decl->GetParameters().GetParams()) {
    if (param->GetID() == "test1" &&
        param->GetParameterType().GetBaseType().ToString() == "int")
      count++;
    if (param->GetID() == "test2" &&
        param->GetParameterType().GetBaseType().ToString() == "int")
      count++;
  }
  EXPECT_EQ(count, 2);

  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetMethodType) {
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", __LINE__,
      tidl::Declaration::MethodType::DELEGATE);
  EXPECT_NE(decl, nullptr);
  EXPECT_EQ(decl->GetMethodType(), tidl::Declaration::MethodType::DELEGATE);
  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetComments) {
  std::string comments = "Test Declaration";
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, comments, __LINE__);
  EXPECT_NE(decl, nullptr);
  EXPECT_EQ(decl->GetComments(), comments);
  delete decl;
}

TEST_F(DeclarationTest, Declaration_GetLine) {
  unsigned line = __LINE__;
  tidl::Declaration* decl = new tidl::Declaration("test",
      new tidl::BaseType("int", ""), params, "", line);
  EXPECT_NE(decl, nullptr);
  EXPECT_EQ(decl->GetLine(), line);
  delete decl;
}

class DeclarationsTest : public testing::Test {
 public:
  tidl::Declaration* decl;

  virtual void SetUp() {}
  virtual void TestDown() {}
  void PrepareDeclaration() {
    tidl::Parameters* params = new tidl::Parameters();
    EXPECT_NE(params, nullptr);
    params->Add(new tidl::Parameter("test1",
          new tidl::ParameterType(new tidl::BaseType("int", "")), __LINE__));
    params->Add(new tidl::Parameter("test2",
          new tidl::ParameterType(new tidl::BaseType("int", "")), __LINE__));
    decl = new tidl::Declaration("test",
        new tidl::BaseType("int", ""), params, "", __LINE__);
    EXPECT_NE(decl, nullptr);
  }
};

TEST_F(DeclarationsTest, Declarations_Constructor) {
  tidl::Declarations* decls = new tidl::Declarations();
  EXPECT_NE(decls, nullptr);
  delete decls;
}

TEST_F(DeclarationsTest, Declarations_Add) {
  tidl::Declarations* decls = new tidl::Declarations();
  EXPECT_NE(decls, nullptr);
  PrepareDeclaration();
  decls->Add(decl);

  bool flag = false;
  for (auto& d : decls->GetDecls()) {
    if (d->GetID() == "test")
      flag = true;
  }
  EXPECT_EQ(flag, true);
  delete decls;
}

TEST_F(DeclarationsTest, Declarations_GetDecls) {
  tidl::Declarations* decls = new tidl::Declarations();
  EXPECT_NE(decls, nullptr);
  PrepareDeclaration();
  decls->Add(decl);

  bool flag = false;
  for (auto& d : decls->GetDecls()) {
    if (d->GetID() == "test")
      flag = true;
  }
  EXPECT_EQ(flag, true);
  delete decls;
}

TEST_F(DeclarationsTest, Declarations_Exist) {
  tidl::Declarations* decls = new tidl::Declarations();
  EXPECT_NE(decls, nullptr);
  PrepareDeclaration();
  decls->Add(decl);
  tidl::Parameters* testParams = new tidl::Parameters();
  EXPECT_NE(testParams, nullptr);
  testParams->Add(new tidl::Parameter("test1",
        new tidl::ParameterType(new tidl::BaseType("int", "")), __LINE__));
  tidl::Declaration* testDecl = new tidl::Declaration("test",
        new tidl::BaseType("int", ""), testParams, "", __LINE__);
  EXPECT_NE(testDecl, nullptr);
  EXPECT_EQ(decls->Exist(testDecl), true);
  delete testDecl;
  delete decls;
}
